;; Generated package description from geiser-chibi.el  -*- no-byte-compile: t -*-
(define-package "geiser-chibi" "0.17.0.20240521.155242" "Chibi Scheme's implementation of the geiser protocols" '((emacs "24.4") (geiser "0.18")) :commit "2502fed1349c2703eea528b74bcc980ad6bceab8" :authors '(("Peter" . "craven@gmx.net")) :maintainer '("Jose A Ortega Ruiz" . "jao@gnu.org") :keywords '("languages" "chibi" "scheme" "geiser") :url "https://gitlab.com/emacs-geiser/chibi")
